// coded by bITmASTER
// bITmASTER@bigfoot.com

#include "libpsx.h"
#include "cdrom.h"
#include "iso9660.h"

#define DIR_LIST 0
#define CD_INT_LIST 0

static volatile char *cd_ctrl = (char *) 0x1f801800;
static volatile int *int_ack = (int *) 0x1f801070;
static volatile int *delay = (int *) 0x1f801018;
static volatile int *dma_ctrl = (int *) 0x1f8010f0;
static volatile int *cd_dma = (int *) 0x1f8010b0;

static int cd_com = 0;
static int cd_blk_size;
static int cd_count;
static int cd_buffer;

static void CdIntDataReady( void );
static void CdIntComplete( void );
static void CdIntAcknowledge( void );
static void CdIntDataEnd( void );
static void CdIntDiskError( void );

volatile int cd_completed = 0;

static char dir_buffer[4096];


char *init_iso9660( void )
{
	ISO_DIRECTORY_RECORD *root_dir;

	CDReadSec( 16, dir_buffer, 1 );								// read Primary Volume Descriptor
	root_dir = (ISO_DIRECTORY_RECORD *) &dir_buffer[156];
	CDReadSec( get_int( root_dir->extent ), dir_buffer, 2 );	// 2 Blcker dir einlesen
	return( dir_buffer );
}


void CDOpen( char *fname, int *start_block, int *size )
{
	int dir_rec_pos;
	ISO_DIRECTORY_RECORD *dir_rec;

	dir_rec_pos = 0;

 	while( dir_rec_pos < 4096 ) {
		dir_rec = (ISO_DIRECTORY_RECORD *) &dir_buffer[dir_rec_pos];
		if ( dir_rec->length[0] == 0 )
			break;													// Directory zu Ende
		dir_rec_pos += dir_rec->length[0];
		if ( ( dir_rec->name[0] == 0 ) || ( dir_rec->name[0] == 1 ) )
			continue;
#if DIR_LIST														// 00 oder 01 
		xprintf( "%s\n", dir_rec->name ); 
#endif
		if ( dir_rec->flags[0] & 1 )
			continue;												// Directory
		if ( strcmp( fname, dir_rec->name ) == 0 ) {
			*start_block = get_int( dir_rec->extent );
			*size = get_int( dir_rec->size );
			return;
		}
	}
 	*start_block = 0;
}	


void CDReadSec( int sec, void *buffer, int size )
{
	char pos[3];

	mmssdd( sec, pos );
	cd_completed = 0;
	SeekCD_( pos );
	while( !cd_completed);
	cd_completed = 0;
	LoadCD_( size, buffer, 0x80 );
	while( !cd_completed);
}

void mmssdd( int block, char *p )
{
	int m, s, d;
	
	block += 150;
	m = block / 4500;			// minuten
	block = block - m * 4500;	// minuten rest
	s = block / 75;				// sekunden
	d = block - s * 75;			// sekunden rest
	
	m = ( ( m / 10 ) << 4 ) | m % 10;
	s = ( ( s / 10 ) << 4 ) | s % 10;
	d = ( ( d / 10 ) << 4 ) | d % 10;	
	
	p[0] = m;
	p[1] = s;
	p[2] = d;
}



//--- CD-ROM Low-Level Routines ---


void SeekCD_( char *pos )
{
	cd_ctrl[0] = 0;
	cd_ctrl[2] = pos[0];			// minute
	cd_ctrl[2] = pos[1];			// second
	cd_ctrl[2] = pos[2];			// sector
	cd_ctrl[1] = 2; 
	cd_com = 0xf2;
}


int LoadCD_( int count, void *buffer, int mode )
{
	
	switch( mode & 0x30 ) {
		case 0x00: cd_blk_size = 512; break;
		case 0x10: cd_blk_size = 582; break;
		case 0x20: cd_blk_size = 585; break;
		case 0x30: cd_blk_size = 0;   break;
	}
	
	if ( !( cd_ctrl[0] & 0x10 ) )
		return( 0 );

	cd_count = count;
	cd_buffer = (int) buffer;
		
	cd_ctrl[0] = 0;
	cd_ctrl[2] = mode;
	cd_com = 0xfe;
	cd_ctrl[1] = 0x0e;
	return( 1 );
}

void cd_int( void )
{
	int status;
	int i;
	
	cd_ctrl[0] = 1;
	status = cd_ctrl[3];
	if ( status == 0 )
		return;
		
	cd_ctrl[0] = 1;
	cd_ctrl[3] = 7;
	
	for ( i = 0; i < 20; i++ );				// etwas warten
	
	cd_ctrl[0] = 1;
	cd_ctrl[3] = status & 0x18;

	for ( i = 0; i < 20; i++ );				// etwas warten

#if CD_INT_LIST
	xprintf( "CDRom-Int %d %x\n", status & 7, cd_com );
#endif
	
	switch( status & 7 ) {
		case 1: CdIntDataReady();
				break;
		case 2: CdIntComplete();
				break;		
		case 3:	CdIntAcknowledge();
				break;
		case 4: CdIntDataEnd();
				break;
		case 5: CdIntDiskError();
				break;
	}		 


	
	*int_ack = ~4;								// Int besttigen
	
}

void CdIntDataReady( void )
{
	if ( cd_com == 0xf6 ) {
		if ( cd_count <= 0 )
			return;
			
		cd_ctrl[0] = 0;
		cd_ctrl[3] = 0;
		cd_ctrl[0] = 0;
		cd_ctrl[3] = 0x80;
		delay[0] = 0x20943;
		delay[2] = 0x132c;
//		dma_ctrl[1] = ( dma_ctrl[1] & 0xffffff ) | 0x880000;	// enable CD-ROM-DMA-Int
		dma_ctrl[0] |= 0x8000;									// enable CD-ROM DMA
		
		cd_dma[0] = cd_buffer;
		cd_dma[1] = cd_blk_size | 0x10000;
		cd_dma[2] = 0x11000000;									// Start DMA
		
		cd_count--;
		cd_buffer += cd_blk_size << 2;
		if ( cd_count > 0 )
			return;

// stop sector reading
	
		cd_ctrl[0] = 0;
		cd_com = 0xfff;
		cd_ctrl[1] = 9;
	}
}

void CdIntComplete( void )
{
	cd_completed = 1;
}

void CdIntAcknowledge( void )
{
	if ( cd_com == 0xf2 ) {
		cd_ctrl[0] = 0;
		cd_ctrl[1] = 0x16;
		cd_com = 0x16;
	}
	
	if ( cd_com == 0xfe ) {
		cd_ctrl[0] = 0;
		cd_ctrl[1] = 6;
		cd_com = 0xf6;
	}
	
}

void CdIntDataEnd( void )
{
}
void CdIntDiskError( void )
{
	xprintf( "CD-Errror !!\n" );
}




int get_int( char *p )
{
	return( ( p[0] & 0xff ) | (( p[1] & 0xff ) << 8 ) | (( p[2] & 0xff ) << 16 ) | (( p[3] & 0xff ) << 24 ) );
}

int get_short( char *p )
{
	return( ( p[0] & 0xff ) | (( p[1] & 0xff ) << 8 ) );
}

